###################################################################################################
# StenaImage
# (C) 2011 Greg Jakes
# http://gjakes.com/project/image-steganographer
#-------------------------------------------------------------------------------------------------- 
# Description: A Python 3.2 image steganographer: hide secret messages within
# 				an image!
#
# Dependencies: Image (PIL module: http://www.pythonware.com/library/pil/)
#
# Command line: python StenaImage.py -enc/-dec [options] "input_filename"
#				-o "filename" : output filename - output image if encrypting, automatically PNG
#												- data if decrypting, no extension added
#				-p "password" : password to use to encrypt/decrypt
#				-d "data" : data to encrypt into image
#				-f "filename" : data from file to encrypt into image
#				-enc : encrypt
#				-dec : decrypt	(either -enc or -dec must be present in command line)
#
# Revisions: v. 1.0 - 11/19/2011 - Initial creation
#
###################################################################################################
# NOTE: Images are only saved as RGBA PNGs in order to retain all pixel data. If the image is
# resized/edited/saved in another format, the encrypted data is most likely lost!
# Unfortunately this also makes for quite a large file, so picture size is limited between
# 10x10 to 2000x2000.
# Possibly a version 2 will be more sophisticated in dealing with this issue!
#
###################################################################################################
# Algorithm Notes:
#	Alters last 2 bits of each RGBA channel, puts them together as GARB to reconstruct byte
#   Writes/checks four corners for char code "$@NE" to verify file and password, put together
#   	as top-left, top-right, bottom-left, bottom-right
#   2nd pixel is used in part to generate encryption/decryption mask
#   3rd pixel from top-left is version byte
#   Next 4 pixels indicate encrypted data length MSB to LSB
#   Data spread evenly through rest of image (VERSION 0x01)
#   First and last rows of image not used for data encryption, in other words, reserved
#
#
#
###################################################################################################

# Imported modules
import Image
import os
import math
from math import floor
import sys
import struct

# Version byte is used to indicate which version of the script encrypted a file,
#	allowing possible algorithm modifications in the future while maintaining
#	backwards compatibility to older files
C_VERSION_BYTE = 0x01

# Char code
C_CHAR_CODE = "$aNE"

# Welcome message
C_WELCOME_MESSAGE = """Welcome to StenaImage v. 1.0!
Created by Greg Jakes 2011 in Python 3.2.2
(Version byte: """ + hex(C_VERSION_BYTE) + """)
--------------------------------------------------------"""

# Usage/Help message
C_USAGE_MESSAGE = """
python StenaImage.py -enc/-dec [options] "input_filename"

**OR IF "StenaImage.exe" IS PRESENT**

StenaImage -enc/-dec [options] "input_filename"

--------------------------------------------------------
OPTIONS:
-o "filename" : output filename - output image if encrypting, automatically PNG
		- data if decrypting, no extension added
-p "password" : password to use to encrypt/decrypt
-d "data" : data to encrypt into image
-f "filename" : data from file to encrypt into image
-enc : encrypt
-dec : decrypt	(either -enc or -dec must be present in command line)"""

# Load image function
def loadimage(filename):
	try:
		print("Attempting to load image: '"+filename+"'")
		im = Image.open(filename)
	except IOError:
		print("Error loading image: "+filename+"!")
		return False
	print("Image loaded, format information: ", im.format, im.size, im.mode)
	
	im = im.convert("RGBA")
	if im.size[0] < 10 or im.size[1] < 10 or im.size[0] > 2000 or im.size[1] > 2000:
		print("Error: Image's dimensions must be between 10x10 and 2000x2000")
		return False
		
	return im
	
# Save image function
def saveimage(image, savefilename):
	try:
		image.save(savefilename)
	except IOError:
		print("Error saving image: "+savefilename+"!")
		return False
	return True
	
# Generate encryption mask function
def genencrmask(password, pixel_1_0):
	pass_len = len(password)
	pass_sum = 0
	for i in range(0, pass_len):
		pass_sum += ord(password[i])
	return [(ord(password[i%pass_len]) + ord(password[(pass_len-i%pass_len)%pass_len]) + int(pass_sum/((i%pass_len)+1)) ^ pixel_1_0[math.ceil(i*1.5)%4])%256 for i in range(0, 256)]
	
# Returns RGBA with encryption mask applied and data in GARB order
def encrybyte(origpixel, databyte, maskbyte):
	encrypteddatabyte = databyte ^ maskbyte
	return (
			((origpixel[0] & 0xFC) | (encrypteddatabyte>>2 & 0x03)),
			((origpixel[1] & 0xFC) | (encrypteddatabyte>>6 & 0x03)),
			((origpixel[2] & 0xFC) | (encrypteddatabyte>>0 & 0x03)),
			((origpixel[3] & 0xFC) | (encrypteddatabyte>>4 & 0x03))
			)
		
# Encrypt function
def stegencrypt(filename, savefilename, password, data):
	im = loadimage(filename)
	if im == False:
		return False
		
	image_width = im.size[0];
	image_height = im.size[1];
	
	data_len = len(data)
	image_capacity = image_width*(image_height-2)
	if data_len > image_capacity:
		print("Error: data length is "+data_len+" bytes, but max data size this image can support is "+image_capacity+" bytes!")
		return False
	print("Data length: %(0)d bytes, Image capacity: %(1)d bytes, Capacity used: %(2)2.3f%%" % {"0":data_len, "1":image_capacity, "2":((data_len/image_capacity)*100)})
	print("Version byte: ", hex(C_VERSION_BYTE))
	
	pixels = im.load()
	mask = genencrmask(password, pixels[1,0])
	#print(mask,"\n")
	
	# Write encrypted char code at 4 corners
	pixels[0, 0] = encrybyte(pixels[0, 0], ord(C_CHAR_CODE[0]), mask[0])
	pixels[image_width-1, 0] = encrybyte(pixels[image_width-1, 0], ord(C_CHAR_CODE[1]), mask[64])
	pixels[0, image_height-1] = encrybyte(pixels[0, image_height-1], ord(C_CHAR_CODE[2]), mask[128])
	pixels[image_width-1, image_height-1] = encrybyte(pixels[image_width-1, image_height-1], ord(C_CHAR_CODE[3]), mask[255])
		
	# Write version byte
	pixels[2, 0] = encrybyte(pixels[2, 0], C_VERSION_BYTE & 0xFF, mask[1])
	
	# Write 4 byte data length
	for i in range(0, 4):
		pixels[i+3, 0] = encrybyte(pixels[i+3, 0], data_len>>(24-i*8) & 0xFF, mask[(i+1)*10])
	
	# Write data
	pixel_ind = range(image_width, image_capacity, int(image_capacity/data_len))
	for i in range(0, data_len):
		pixels[pixel_ind[i]%image_width, pixel_ind[i]/image_width] = encrybyte(pixels[pixel_ind[i]%image_width, pixel_ind[i]/image_width], data[i], mask[i%256])
	
	print("Attempting to save new image as '"+savefilename+"'")
	if saveimage(im, savefilename) == False:
		return False
	
	return True

# Returns data byte from encrypted RGBA pixel
def decrybyte(origpixel, maskbyte):
	return   ((origpixel[1] & 0x03)<<6
			| (origpixel[3] & 0x03)<<4
			| (origpixel[0] & 0x03)<<2
			| (origpixel[2] & 0x03)<<0
			  ) ^ maskbyte
			
	
# Decrypt function
def stegdecrypt(filename, password):
	im = loadimage(filename)
	if im == False:
		return False
		
	image_width = im.size[0];
	image_height = im.size[1];
	
	image_capacity = image_width*(image_height-2)
	
	pixels = im.load()
	mask = genencrmask(password, pixels[1,0])
	
	# Get and check char code at 4 corners
	char_code = chr(decrybyte(pixels[0, 0], mask[0])) + chr(decrybyte(pixels[image_width-1, 0], mask[64])) + chr(decrybyte(pixels[0, image_height-1], mask[128])) + chr(decrybyte(pixels[image_width-1, image_height-1], mask[255]))
				
	#print(str(char_code))
	if str(char_code) != C_CHAR_CODE:
		print("Error validating char code, this most likely means you tried the wrong password or that the image was corrupted/edited.")
		return False
	
	# Read version byte
	version_byte = decrybyte(pixels[2, 0], mask[1])
	print("Version byte: ", hex(version_byte))
	
	# Get 4 byte data length
	data_len = 0
	for i in range(0, 4):
		data_len = data_len<<8 | decrybyte(pixels[i+3, 0], mask[(i+1)*10])
	if data_len > image_capacity:
		print("Error: data length was read as "+data_len+" bytes, but max data size this image can support is "+image_capacity+" bytes! This means the image was corrupted or resized.")
		return False
	print("Data length: %(0)d bytes, Image capacity: %(1)d bytes, Capacity used: %(2)2.3f%%" % {"0":data_len, "1":image_capacity, "2":((data_len/image_capacity)*100)})	
	
	# Read data
	pixel_ind = range(image_width, image_capacity, int(image_capacity/data_len))
	data = [decrybyte(pixels[pixel_ind[i]%image_width, pixel_ind[i]/image_width], mask[i%256]) for i in range(0, data_len)]
	
	return data
	
# Main
print(C_WELCOME_MESSAGE)

# Set defaults
in_filename = ""
out_filename = ""
password = "password"
data = "data"
args_len = len(sys.argv)

if args_len > 2:
	in_filename = sys.argv[args_len-1]
	enc_flag = 0
	if sys.argv[1] == "-enc":
		enc_flag = 1
		out_filename = os.path.splitext(in_filename)[0] + "_enc.png"
	elif sys.argv[1] == "-dec":
		enc_flag = 2
	else:
		print("Command-line Error: -enc or -dec must be the first argument!")
		
	if enc_flag > 0:	
		for i in range(2, args_len-1):
			if sys.argv[i] == "-o" and i < args_len-2:
				if enc_flag == 1:
					out_filename = os.path.splitext(sys.argv[i+1])[0] + ".png"
				else:
					out_filename = sys.argv[i+1]
			elif sys.argv[i] == "-p" and i < args_len-2:
				password = sys.argv[i+1]
			elif sys.argv[i] == "-d" and data == "data" and i < args_len-2:
				data = [ord(c) for c in sys.argv[i+1]]
			elif sys.argv[i] == "-f" and data == "data" and i < args_len-2:
				print("Attempting to load data from file '"+sys.argv[i+1]+"'")
				datafile = open(sys.argv[i+1], "rb")
				data = struct.unpack(str(os.path.getsize(sys.argv[i+1]))+"B", datafile.read(os.path.getsize(sys.argv[i+1])))
			elif sys.argv[i][0] == '-':
				print("Command-line Warning: '"+sys.argv[i]+"' argument ignored!")
		
		if enc_flag == 1:
			if stegencrypt(in_filename, out_filename, password, data):
				print("Success encrypting!")
			else:
				print("Failure encrypting!")
				
		elif enc_flag == 2:	
			print("\nNow attempting to decrypt...")
			data = stegdecrypt(in_filename, password)
			if not data:
				print("Failure decrypting!")
			else:
				print("Success decrypting!")
				if out_filename == "":
					print("Data: ", "".join(map(chr, data)))
				else:
					print("Data being written to '"+out_filename+"'")
					datafile = open(out_filename, "wb")
					for d in data:
						datafile.write(struct.pack("B",d))
			
		
	
else:
	print("Command-line Error:\nUsage:")
	print(C_USAGE_MESSAGE)
	
	
print("")




